from ._frame import select,subset
from ._IO import saveimg
from .utils import dict_slice,flatten,unique_exprs,get_TCGA_mRNA
from .gene import current_symbol,deg_siglabel,deg_filter,deseq,geneIDconverter,countto

__all__=()


import pandas as pd
import numpy as np
import matplotlib.pyplot as plt


def check_axes(fig, ax, fig_shape=None):
    if all([x is None for x in [fig, ax]]):
        fig_shape = (8, 19) if fig_shape is None else None
        fig, ax = plt.subplots(figsize=fig_shape)

    return fig, ax


def get_xlim(ax):
    xmin, xmax = ax.get_xaxis().get_view_interval()

    return xmin, xmax


def get_ylim(ax):
    ymin, ymax = ax.get_yaxis().get_view_interval()

    return ymin, ymax


def save_fig(fig, file_name, save_type='pdf', dpi=500):
    if save_type == 'pdf':
        fig.savefig('{}.pdf'.format(file_name), transparent=False, dpi=dpi,
                    bbox_inches='tight')

    elif save_type == 'tiff':
        fig.savefig('{}.tiff'.format(file_name), transparent=False, dpi=dpi,
                    bbox_inches='tight', pil_kwargs={'compression': 'tiff_lzw'})
    else:
        raise ValueError('Figure type not supported')


def check_data(data):
    assert isinstance(
        data, pd.DataFrame), 'Please supply data as Pandas DataFrame'

    data['sort_order'] = np.arange(data.shape[0])[::-1]

    return data


def forest_plot(data, auc_col, se_col, to_annotate=None, subset_col=None, fig=None, ax=None, add_legend=True,
                legend_loc='upper right', hbar_lim=None, xlim=None, fig_shape=None, anot_base=0.15,
                annot_offset=13, annot_scaler=0.01, add_ci=True, auc_label=None, fargs=None
                ):
    """
    Basic forest plots for meta analyses

    Parameters
    ----------
    data: pandas DataFrame
        contains all columns to be annotated, AUCs and standard errors
    auc_col: str
        named of column containing AUC values
    se_col: str
        name of column containing standard errors
    to_annotate: list of str, default None
        column names for information to annotation on the left of the plot
    subset_col: str, default None
        name of column for splitting data by marker type in plot
    fig: matplotlib figure, default None
        use if matplotlib figure/axes have already been created
    ax: matplotlib axes, default None
        use if matplotlib figure/axes have already been created
    add_legend: bool, default True
        add a legend
    legend_loc: str, default 'upper right'
        location of legend
    hbar_lim: tuple or list, default None
        x-axis limits for the header/tail of table
    xlim: tuple or list, default None
        x-axis limits for plotting AUC and error bars
    fig_shape: tuple, default None
        figure size to use if you want to overwrite the default shape
    anot_base: float, default 0.15
        minimum distance between all columns, regardless of column width
    annot_offset: float, default 13
        estimate of average number of characters per column. Adjust up or down accordingly
    annot_scaler: float, default 0.01
        estimate of the amount by extra/reduced space to give column widths that deviate from annot_offset
    add_ci: bool, default True
        Annotate CI after AUC
    auc_label: str or None, default None
        Label for the forest plot section
    fargs: dict, default None
         kwargs for fontsizes

    Returns
    -------
    fig, ax: matplotlib figure and axes objects
    """
    data = check_data(data)
    fig, ax = check_axes(fig, ax, fig_shape)

    # plot auc and 95% CI
    ax = plot_auc(data, ax, auc_col, se_col, subset_col=subset_col, add_legend=add_legend, legend_loc=legend_loc,
                  xlim=xlim, auc_label=auc_label)

    # add all columns in to_annotate as text
    if to_annotate is not None:
        ax = annotate_columns(data, to_annotate, auc_col, se_col, ax, anot_base=anot_base, annot_offset=annot_offset,
                              annot_scaler=annot_scaler, hbar_lim=hbar_lim, add_ci=add_ci, auc_label=auc_label,
                              fargs=fargs)

    return fig, ax


def parse_auc(data, row, auc, se_col, add_ci=True):
    if add_ci:
        auc = '{:.2f} [{:.2f}-{:.2f}]'.format(auc, auc - (1.96 * data.iloc[row, :][se_col]),
                                              auc + (1.96 * data.iloc[row, :][se_col]))
    else:
        auc = '{:.2f}'.format(auc)

    return auc


def annotate_columns(data, to_annotate, auc_col, se_col, ax, anot_base=0.15, annot_offset=13, annot_scaler=0.01,
                     y_offset=0.25, hbar_lim=None, add_ci=True, auc_label=None, fargs=None):
    xmin, xmax = get_xlim(ax)
    fargs = {'fontsize': 12, 'fontfamily': 'sans-serif',
             'clip_on': False} if fargs is None else fargs
    to_annotate = [to_annotate] if isinstance(
        to_annotate, str) else to_annotate
    assert isinstance(
        to_annotate, list), 'supply annotation column names as a list of strings'
    to_annotate = to_annotate[::-1]
    column_offsets = []
    annotations = []

    for i, txt in enumerate(np.arange(data.shape[0])):
        xiter = xmin  # moves column annotation outward from plot
        for j, column in enumerate(to_annotate):
            max_len = max(data[column].astype(str).str.len().max(), len(
                column))  # gets width of a column in characters
            col_shift = anot_base + ((max_len-annot_offset)*annot_scaler)
            if j != 0:
                # distance away from previous shifted proportional to number of characters
                xiter -= col_shift
            s = data.iloc[txt, :][column]
            if isinstance(s, (float, np.float64)) and np.isnan(s):
                continue
            s = parse_auc(data, txt, s, se_col,
                          add_ci) if column == auc_col else s
            x = xmax if column == auc_col else xiter
            y = data.iloc[txt, :]['sort_order'] - y_offset
            annotations.append(ax.text(x=x, y=y, s=s, **fargs))
            if i == 0:
                auc_label = 'AUC' if auc_label is None else auc_label
                annotations.append(
                    ax.text(x=x, y=data.shape[0]+0.6, s=column, **fargs))
                if add_ci:
                    __ = [annotations.append(ax.text(x=aucx, y=data.shape[0]+0.6, s=auc_label + ' [95%  CI]', **fargs))
                          for aucx in [xmin, xmax]]
                else:
                    __ = [annotations.append(ax.text(x=aucx, y=data.shape[0]+0.6, s=auc_label, **fargs))
                          for aucx in [xmin, xmax]]
                column_offsets.append(col_shift)

    bars = []
    annotations = [a.get_position()[0] for a in annotations]
    if hbar_lim is None:
        #hbar_xmin = 0 - xmin - np.sum(column_offsets) + column_offsets[0]
        #hbar_xmax = 1.22
        if add_ci:
            max_bar = np.max(annotations) if auc_label is None else np.max(
                annotations) + 0.009 * len(auc_label + ' [95%  CI]')
        else:
            max_bar = np.max(annotations) if auc_label is None else np.max(
                annotations) + 0.01 * len(auc_label)
        hbar_xmin, hbar_xmax = np.min(annotations), max_bar

    else:
        hbar_xmin, hbar_xmax = hbar_lim

    bargs = {'c': '#333333', 'linestyle': '-'}
    for hbar_ypos, line_width in zip([data.shape[0], data.shape[0]+2, -2.8], [1, 2, 2]):
        bars.append(ax.axhline(y=hbar_ypos, xmin=hbar_xmin,
                    xmax=hbar_xmax, linewidth=line_width, **bargs))
    _ = [b.set_clip_on(False) for b in bars]
    _ = [b.set_transform(ax.transData) for b in bars]

    return ax


def plot_auc(data, ax, auc_col, se_col, subset_col=None, colour='#2E5A86', add_vline=1, add_legend=True,
             legend_loc='upper right', xlim=None, auc_label=None):
    if subset_col is not None:
        subsets = data[subset_col].unique().tolist(
        ) if subset_col in data.columns else None
    else:
        subsets = None

    if subsets is not None:
        markers = ['o', 's', '^', 'v', 'X', 'P', 'p'][:len(subsets)]
        assert len(markers) == len(
            subsets), 'More subsets than markers - overwrite marker list'

        for s, m in zip(subsets, markers):
            ax.errorbar(data.loc[data[subset_col] == s, auc_col],
                        data.loc[data[subset_col] == s, 'sort_order'],
                        xerr=data.loc[data[subset_col] == s, se_col] * 1.96,
                        fmt=m, label=s, c=colour, elinewidth=0.8,
                        **{'markersize': 7})
    else:
        auc_label = auc_label if auc_label is not None else 'AUC (95% CI)'
        ax.errorbar(data[auc_col],
                    data['sort_order'],
                    xerr=data[se_col] * 1.96,
                    fmt='o', label=auc_label, c=colour, elinewidth=0.8,
                    **{'markersize': 7})

    if add_vline:
        ax.axvline(x=add_vline, c='#474747', linestyle='--', alpha=0.8, ymin=0,
                   ymax=0.99 * ((data.shape[0] - 1) / data.shape[0]))

    if add_legend:
        leg = ax.legend(loc=legend_loc, frameon=False, title="Validation", fontsize=11,
                        borderaxespad=2.2, framealpha=0.5)
        leg._legend_box.align = "left"
    ax.tick_params(axis='x', labelsize=12)
    ax.set_frame_on(False)
    ax.set_yticks([])
    ax.set_ylim([-1, data.shape[0]+1])
    if xlim is not None:
        ax.set_xlim(*xlim)
    xmin, xmax = get_xlim(ax)
    ymin, ymax = get_ylim(ax)
    ax.add_artist(plt.Line2D((xmin, xmax), (ymin, ymin),
                  color='#262626', linewidth=2))
    ax.tick_params(direction="in", which='major', pad=3, length=7)
    ax.tick_params(direction="in", which='minor', pad=3, length=4)
    ax.minorticks_on()

    return ax



from ..st import detects
import pandas as pd
from typing import Union, Optional


def select(frame, columns=None, pattern=None):
    """
    select a DataFrame columns according to `subsets` conditions
    """
    _objs = []
    if columns:
        cidx = frame.columns.isin(values=columns)
        _objs.append(frame.loc[:, cidx])
    if pattern:
        cidx = detects(string=frame.columns, pattern=pattern)
        _objs.append(frame.loc[:, cidx])
    return pd.concat(objs=_objs,axis=1)


def subset(
    frame: pd.DataFrame,
    subsets: dict,
    inplace: bool = False
) -> Optional[pd.DataFrame]:
    """
    filter/subset a DataFrame according to `subsets` conditions
    """
    _f = frame if inplace else frame.copy()
    for k in subsets:
        v = subsets.get(k)
        if isinstance(v, list):
            _lg = _f[k].isin(v)
            _f = _f.loc[_lg, :]
        else:
            _lg = _f[k].apply(lambda x: eval(v))
            _f = _f.loc[_lg, :]
    return None if inplace else _f



import numpy as np
import re

def sub(string,start=0,stop=-1) -> str:
    return string[start:stop]

def detect(pattern: str, string: str, flags=0) -> bool:
    return False if re.search(pattern=pattern,string=string,flags=flags) is None else True

def replace(pattern: str, repl: str, string: str, count:int = 0):
    return re.sub(pattern=pattern,repl=repl,string=string, count=count)

def grep(pattern: str, string: str, flags=0):
    bl = detect(pattern=pattern,string=string,flags=flags)
    return list(np.array(string)[bl])

def count(pattern: str, string: str, flags=0):
    return len(re.findall(pattern=pattern,string=string,flags=flags))


import numpy as np
import pandas as pd
from ._frame import subset,select
from ._IO import read_csv_gz
import atopos
import re
import functools

def current_symbol(frame,reference,tax_id=9606):
    '''
    map SYMBOL alias to latest official SYMBOL NAME
    frame: dataframe,index is SYMBOL
    reference: dataframe, can download from ...
    '''
    alias=pd.read_feather(reference)
    alias = subset(alias,{"tax_id":[tax_id]})
    alias = select(alias,columns=["Symbol","Alias"])
    alias.set_index(keys="Symbol",drop=True,inplace=True)
    lg = frame.index.isin(alias.index)
    new = pd.merge(frame.loc[~lg,:],alias,left_index=True,right_on="Alias",how='left')
    
    return pd.concat([new,frame.loc[lg,:]])

def deg_siglabel(
        table: pd.DataFrame, 
        lfc='LFC',
        pvalue:str = "Pvalue", # "Padj"
        lfc_thr = (.585, .585),
        pv_thr =(.05, .05),
        siglabel=('Down', 'NS', 'Up'),
        inplace=True
        ) -> pd.DataFrame:
    """
    label genes for significant up/down or not significant
    lfc
    pvalue
    """
    # upregulated
    _table = table if inplace else table.copy()
    lg_up = np.logical_and(_table[lfc] >= lfc_thr[1],_table[pvalue] < pv_thr[1])
    _table.loc[lg_up,'Change'] = siglabel[2]
    # downregulated
    lg_down = np.logical_and(_table[lfc] <= -lfc_thr[0],_table[pvalue] < pv_thr[0])
    _table.loc[lg_down, 'Change'] = siglabel[0]
    _table.fillna(value={'Change': siglabel[1]}, inplace=True)
    # return _table
    print(f'All degs: {_table.loc[_table.Change != siglabel[1], :].shape[0]}')
    print(
        f'Up: {_table.loc[_table.Change == siglabel[2],:].shape[0]}')
    print(
        f'Down: {_table.loc[_table.Change == siglabel[0],:].shape[0]}')
    return None if inplace else _table

def deg_filter(
        frame: pd.DataFrame,
        lfc :str='LFC',
        top_n=None,
        filter_label=['Up','Down']
        ) -> pd.DataFrame:
    if top_n:
        _df = frame.sort_values(by=lfc)
        nrow = frame.shape[0]
        dfslice = list(range(0, top_n)) + \
            list(range(nrow-top_n, nrow))
        return _df.iloc[dfslice, :]
    else:
        return subset(frame,{"Change":filter_label})

def deseq(
        fdata: pd.DataFrame,
        pdata: pd.DataFrame,
        case: str,
        ref: str,
        n_jobs: int = 2) -> pd.DataFrame:
    """
    differential expression analysis (DEA) with bulk RNA-seq data
    fdata:pandas.DataFrame One column per gene, rows are indexed by sample barcodes.
    pdata:pandas.DataFrame, first column is Group
    """
    import pydeseq2
    from pydeseq2.dds import DeseqDataSet
    from pydeseq2.ds import DeseqStats
    rename = functools.partial(re.sub,pattern="_",repl="-",count=0)
    renames = np.vectorize(rename,otypes=[str])
    case=rename(string=case)
    ref=rename(string=ref)
    pdata["Condition"]=renames(string=pdata.iloc[:,0])
    pdata = pdata.loc[pdata.Condition.isin([ref,case]),:]
    common_sample = np.intersect1d(fdata.index,pdata.index)
    fdata = fdata.loc[common_sample,:]
    pdata = pdata.loc[common_sample,:]
    pdata["Condition"] = pdata.Condition.astype('category').values
    genes_to_keep = fdata.columns[fdata.sum(axis=0) >= 10]
    fdata = fdata[genes_to_keep]
    inference=pydeseq2.default_inference.DefaultInference(joblib_verbosity=0, n_cpus=n_jobs)
    dds = DeseqDataSet(
        counts=fdata,
        metadata=pdata,
        ref_level=["Condition", ref],
        design_factors="Condition",
        refit_cooks=True,
        inference=inference
    )
    # 离散度和log fold-change评估.
    dds.deseq2()
    # 差异表达统计检验分析
    stat_res = DeseqStats(
        dds,
        alpha=0.05,
        cooks_filter=True,
        contrast=['Condition', case, ref],
        independent_filter=True,
        inference=inference
    )
    stat_res.summary()
    return stat_res.results_df.reset_index().rename(
        columns=dict(
            Gene="Feature", log2FoldChange="LFC", padj="Padj", pvalue="Pvalue"
        )
    )
    

def geneIDconverter(frame, from_id='Ensembl', to_id='Symbol',species="hsa", keep_from=False, gene_type=None):
    if species == "hsa":
        file_path = "h38_gene_info_v43.csv.gz"
    if species == "mmu":
        file_path = "m39_gene_info_v32.csv.gz"
    annot = _IO.read_csv_gz(file_path)
    if from_id=='Ensembl':
        annot.loc[:,"Ensembl"] = atopos.st.removes(string=annot.Ensembl,pattern=r"\.\d+")
    annot.set_index(keys=from_id,inplace=True,drop=True)
    if gene_type:
        gene_type = annot.GeneType.isin(gene_type)
        annot = annot.loc[gene_type:, [to_id]]
    else:
        annot = annot.loc[:, [to_id]]
    if keep_from:
        annoted = pd.merge(annot, frame, left_index=True, right_index=True)
    else:
        annoted = pd.merge(annot, frame, left_index=True, right_index=True)
        annoted.set_index(keys=to_id, inplace=True)
    return annoted

def countto(frame, towhat="tpm", geneid='Ensembl', species="hsa"):
    '''
    towhat: tpm(default), fpkm, cpm
    return: a dataframe
    '''
    from bioinfokit import analys
    if species == "hsa":
        file_path = "h38_gene_info_v43.csv.gz"
    if species == "mmu":
        file_path = "m39_gene_info_v32.csv.gz"
    annot = read_csv_gz(file_path, usecols=[
                        geneid, 'Length'])
    if geneid=='Ensembl':
        annot.loc[:,"Ensembl"] = atopos.st.removes(string=annot.Ensembl,pattern=r"\.\d+")
    annot.set_index(keys=geneid,inplace=True,drop=True)
    _df = pd.merge(annot, frame, left_index=True, right_index=True)
    nm = analys.norm()
    nm.tpm(df=_df, gl='Length')
    nm.rpkm(df=_df, gl='Length')
    nm.cpm(df=_df)
    if towhat=="tpm":
        return nm.tpm_norm
    if towhat=="fpkm":
        return nm.rpkm_norm
    if towhat=="cpm":
        return nm.cpm_norm.drop(columns="Length")



import functools
import gzip
from typing import Tuple, Union

import matplotlib.pyplot as plt
import pandas as pd
from importlib import resources
from typing import Union
import pathlib

from .. import _data


def _output(filename: str, _format: str, *, outdir: Union[pathlib.PosixPath,str] = pathlib.Path().absolute()) -> pathlib.PosixPath:
    if isinstance(outdir,str):
        return pathlib.Path(outdir).joinpath(f"{filename}.{_format}")
    elif isinstance(outdir,pathlib.PosixPath):
        return outdir.joinpath(f"{filename}.{_format}")
    else:
        raise ValueError(f'{outdir} should be string or pathlib.PosixPath class')


def _saveimg(
    filename: str,
    formats: Union[str, Tuple[str, ...]],
    *,
    outdir: Union[pathlib.PosixPath,str] = pathlib.Path().absolute(),
    dpi: int = 300,
    figsize=(7, 7),
) -> None:
    if isinstance(formats, str):
        formats = (formats,)

    plt.rcParams["figure.figsize"] = figsize
    for i in formats:
        plt.savefig(
            _output(filename=filename, _format=i, outdir=outdir),
            dpi=dpi,
            bbox_inches="tight",
        )
    plt.close()


def saveimg(
    formats: Union[str, Tuple[str, ...]],
    *,
    outdir: Union[pathlib.PosixPath,str] = pathlib.Path().absolute(),
    dpi: int = 300,
    figsize=(7, 7),
):
    return functools.partial(
        _saveimg, formats=formats, outdir=outdir, dpi=dpi, figsize=figsize
    )


def mkdir(dir: str, parents=True, exist_ok=True):
    pathlib.Path(dir).mkdir(parents=parents, exist_ok=exist_ok)


def read_csv_gz(
    data_file_name: str,
    index_col=None,
    usecols=None,
    sep=",",
    *,
    encoding="utf-8",
    **kwargs,
):
    """Loads gzip-compressed with `importlib.resources`.

    1) Open resource file with `importlib.resources.open_binary`
    2) Decompress file obj with `gzip.open`
    3) Load decompressed data with `pd.read_csv`

    Parameters
    ----------
    data_file_name : str
            Name of gzip-compressed csv file  (`'*.csv.gz'`) to be loaded from
            `_data/data_file_name`. For example `'humanGene.csv.gz'`.
    """

    with resources.open_binary(_data, data_file_name) as _:
        _ = gzip.open(_, mode="rt", encoding=encoding)
        _df = pd.read_csv(_, usecols=usecols, index_col=index_col, sep=sep)
    return _df



import functools
import numpy as np
from ._stringpy import *

# string
subs = np.vectorize(sub,otypes=[str])
replaces = np.vectorize(replace,otypes=[str])
detects = np.vectorize(detect,otypes=[bool])
remove = functools.partial(replace,repl='')
removes = np.vectorize(remove,otypes=[str])
counts = np.vectorize(count,otypes=[int])

def greps(pattern: str, string: str, flags=0):
    bl = detects(pattern=pattern,string=string,flags=flags)
    return list(np.array(string)[bl])


import itertools
import re
import numpy as np
import pandas as pd
from ..st import removes,subs
from ._frame import select


# flatten a nest list
# flatten = lambda nest_list: sum(([x] if not isinstance(x, list) else flatten(x) for x in nest_list), [])

def dict_slice(adict:dict, start:int, end:int)->dict:
    '''
    slice a dict using numeric index
    '''
    keys = adict.keys()
    dict_slice = {}
    for k in list(keys)[start:end]:
        dict_slice[k] = adict[k]
    return dict_slice

def flatten(nest_list):
    nl = nest_list.copy()
    import itertools
    for x,y in enumerate(nl):
        if not isinstance(y, list):
            nl[x] = [y]
    return list(itertools.chain.from_iterable(nl))


def unique_exprs(frame, reductions=np.median):
    """
    基因去重复
    frame: row_index = genenames, columns = samples
    """
    frame['Ref'] = frame.apply(reductions, axis=1)
    frame.sort_values(by='Ref', ascending=False, inplace=True)
    frame.drop(columns='Ref', inplace=True)
    frame['Ref'] = frame.index
    frame.drop_duplicates(subset='Ref', inplace=True)
    return frame.drop(columns='Ref')

def get_TCGA_mRNA(arrow, dtype='tpm', label="01A", gene_type:list=None, barcode_length=16):
    _df = pd.read_feather(arrow)
    _df.loc[:,"gene_id"] = removes(string=_df.gene_id,pattern=r"\.\d+")
    _df.set_index(keys="gene_id", drop=True, inplace=True)

    # 筛选编码蛋白基因
    if gene_type:
        _df = _df.loc[_df.gene_type.isin(gene_type), :]
    # 筛选数据
    _df = select(frame=_df,columns=["gene_name","gene_type"], pattern=re.compile(pattern=f"^{dtype}_"))

    _df.columns = removes(
        string=_df.columns.values, pattern=re.compile(pattern=f"^{dtype}_"))

    _df = select(frame=_df,columns=["gene_name","gene_type"], pattern=re.compile(pattern=f"^.{{13}}{label}"))
    # barcode_length of samples
    _df.columns = subs(string=_df.columns.values, start=0, stop=barcode_length)

    return _df




import numpy as np
import pathlib
import pandas as pd
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from ..pl._palette import Palette
from ..tl._IO import saveimg
from typing import Tuple, Union

def ROC(y_true:dict, y_hat:dict,xlabel=None,ylabel=None, outdir=None,filename=''):
    """
    plot multiple ROC curves in one figure
    y_true: dict
    y_hat: dict
    outdir: output directory
    filename: filename
    """
    from sklearn.metrics import RocCurveDisplay
    colors = dict(zip(y_hat,Palette.set2))
    for i in y_hat.keys():
        RocCurveDisplay.from_predictions(y_true[i], y_hat[i], name = i, color= colors[i], ax=plt.gca(),linewidth=2)
    plt.plot([0,1],[0,1], linestyle="dashed",color = "grey");
    if ylabel:
        plt.ylabel(ylabel)
    else:
        plt.ylabel("True Positive Rate");
    if xlabel:
        plt.xlabel(xlabel);
    else:
        plt.xlabel("False Positive Rate");
    p = plt.gcf()
    p.set_size_inches(6, 6)
    plt.close()
    if outdir:
        p.savefig(pathlib(outdir).joinpath(f"{filename}.pdf"))
    return p

def CC(
    y_true:dict,
    y_hat:dict,
    outdir: Union[pathlib.PosixPath,str] = pathlib.Path().absolute(),
    filename:str='CalibrationCurve') -> Figure:
    """
    y_true: dict
    y_hat: dict
    outdir: output directory
    filename: filename
    """
    from sklearn.calibration import CalibrationDisplay
    from sklearn.preprocessing import MinMaxScaler
    colors = dict(zip(y_hat,Palette.set2))
    if any(y_hat.min()<0) or any(y_hat.max()>0):
        y = MinMaxScaler((.0000000001,.9999999999)).fit_transform(y_hat)
    else:
        y = y_hat.values
    for i in y_hat.keys():
        CalibrationDisplay.from_predictions(y_true[i], y_hat[i], name = i, color= colors[i], ax=plt.gca(),linewidth=2)
    plt.plot([0,1],[0,1], linestyle="dashed",color = "grey");
    plt.ylabel("True Positive Rate");
    plt.xlabel("False Positive Rate");
    p = plt.gcf()
    p.set_size_inches(6, 6)
    plt.close()
    if outdir:
        p.savefig(pathlib(outdir).joinpath(f"{filename}.pdf"))
    return p

# def learning_curve(x,xlabel='',ylabel='',mark:int=0, outdir:str=None) -> Figure :
#     plt.rc('axes', labelsize=16) #fontsize of the x and y labels
#     plt.rc('xtick', labelsize=12) #fontsize of the x tick labels
#     plt.rc('ytick', labelsize=12)
#     plt.plot(np.arange(mark, len(x)+mark), 
#         x,
#         color='grey', 
#         linewidth=.8,
#         # linestyle='dashed', 
#         marker='o',
#         markerfacecolor='white',
#         markeredgecolor='r',
#         markersize=5
#         )
#     plt.plot(np.where(x==x.max())[0] + mark,
#         x.max(),
#         marker='o',
#         markerfacecolor='white',
#         markersize=10,
#         markeredgecolor='b'
#         )
#     plt.gca().spines.right.set_visible(False)
#     plt.gca().spines.top.set_visible(False)
#     plt.ylabel(ylabel)
#     plt.xlabel(xlabel)
#     p = plt.gcf()
#     p.set_size_inches(9, 6)
#     plt.close()
#     if outdir:
#         p.savefig(outdir + "/LearningCurve"+ ".pdf",dpi=300)
#     return p
